@@ -1,17 +1,31 @@
 mefp <- function(obj, ...) UseMethod("mefp")
 
 mefp.formula <-
-    function(formula, data=list(), type = c("ME", "fluctuation"), h=1,
-             alpha=0.05, functional = c("max", "range"),
+    function(formula, type = c("OLS-CUSUM", "OLS-MOSUM", "ME", "fluctuation"),
+             data=list(), h=1, alpha=0.05, functional = c("max", "range"),
              period=10, tolerance=.Machine$double.eps^0.5,
-             MECritvalTable=monitorMECritvalTable, rescale=FALSE, ...)
+             CritvalTable=NULL, rescale=NULL, border=NULL, ...)
 {
     type <- match.arg(type)
     functional <- match.arg(functional)
-    val <- efp(formula, type=type, h=h, data=data, rescale=rescale)
+
+    histrescale <- rescale
+    if(type=="fluctuation") histrescale <- TRUE
+    if(is.null(rescale)) {
+      if(type=="fluctuation") {
+        rescale <- FALSE
+	histrescale <- TRUE
+      }
+      else {
+        rescale <- TRUE
+	histrescale <- TRUE
+      }
+    }
+
+    val <- efp(formula, type=type, h=h, data=data, rescale=histrescale)
     val <- mefp(val, alpha=alpha, functional=functional, period=period,
-                tolerance=tolerance, MECritvalTable=MECritvalTable,
-                rescale=rescale)
+                tolerance=tolerance, CritvalTable=CritvalTable,
+                rescale=rescale, border=border, ...)
     if(length(data) == 0)
         val$data <- NULL
     else
@@ -23,11 +37,11 @@ mefp.formula <-
 mefp.efp <-
     function(obj, alpha=0.05, functional = c("max", "range"),
              period=10, tolerance=.Machine$double.eps^0.5,
-             MECritvalTable=monitorMECritvalTable, rescale=NULL, ...)
+             CritvalTable=NULL, rescale=NULL, border=NULL, ...)
 {
     functional <- match.arg(functional)
-    if(! (obj$type %in% c("ME", "fluctuation")))
-        stop("efp must be of type `fluctuation' or `ME'")
+    if(! (obj$type %in% c("OLS-CUSUM", "OLS-MOSUM", "ME", "fluctuation")))
+        stop("efp must be of type `OLS-CUSUM', `OLS-MOSUM', `fluctuation' or `ME'")
 
     if(is.null(as.list(obj$call)$data)){
        data <- NULL
@@ -36,30 +50,53 @@ mefp.efp <-
        data <- as.character(as.list(obj$call)$data)
     }
 
-    if(is.null(rescale)) rescale <- as.list(obj$call)$rescale
+    if(is.null(rescale) & (obj$type == "fluctuation")) rescale <- FALSE
+    if(is.null(rescale)) rescale <- obj$rescale
     if(is.null(rescale)) rescale <- TRUE
 
     ## Bonferroni correction
     elemsiglevel <- alpha / obj$nreg
     histcoef <- obj$coefficients
     histsize <- obj$nobs
-    switch(obj$type,
-           "ME" = { winsize <- obj$par },
-       { winsize <- 1 })
+    winsize <- obj$par
     K <- floor(winsize*obj$nobs)
     sigmahat <- obj$sigma
-    Q12s <- obj$Q12 * K/(sigmahat*sqrt(histsize))
-
-    computeEmpProc <- function(newcoef, Q){
-        if(is.null(Q)) Q <- Q12s
-        else Q <- Q * K/(sigmahat*sqrt(histsize))
-        t(Q %*%(newcoef-histcoef))
-    }
+    Q12 <- obj$Q12
 
     logPlus <- function(x) ifelse(x<=exp(1),1,log(x))
 
-    if(obj$type=="ME"){
-        dntab <- dimnames(MECritvalTable)
+    switch(obj$type,
+
+    "OLS-CUSUM" = {
+
+        mreSize <- function(a){
+            -2*(pnorm(a)-a*dnorm(a))
+        }
+        mreCritval <- function(a){
+            abs(2*(pnorm(a)-a*dnorm(a))+alpha-2)
+        }
+        critval <- optim(5, mreCritval)$par
+        if((mreSize(critval)-alpha) > tolerance)
+            stop("Could not find critical within tolerance")
+
+        computeEmpProc <- function(X, y)
+        {
+          as.vector(cumsum((y - X %*% histcoef)/(sigmahat*sqrt(histsize))))
+        }
+        computeEstims <- NULL
+        functional <- "max"
+        if(is.null(border)){
+            border <- function(k){
+                x <- k/histsize
+                sqrt(x*(x-1)*(critval^2 + log(x/(x-1))))
+            }
+        }
+    },
+
+    "OLS-MOSUM" = {
+        if(is.null(CritvalTable))
+            CritvalTable <- get("monitorMECritvalTable")
+        dntab <- dimnames(CritvalTable)
         if(!(winsize %in% dntab[[1]]))
             stop(paste("winsize h =",winsize,"not available, we have:",
                        paste(dntab[[1]], collapse=", ")))
@@ -67,7 +104,47 @@ mefp.efp <-
             stop(paste("period",period,"not available, we have:",
                        paste(dntab[[2]], collapse=", ")))
         critval <- approx(x=as.numeric(dntab[[3]]),
-                          y=MECritvalTable[as.character(winsize),
+                          y=CritvalTable[as.character(winsize),
+                          as.character(period),,functional],
+                          xout=1-alpha)$y
+        if(is.na(critval))
+            stop(paste("Necessary significance level per parameter of",
+                       alpha,
+                       "\n\toutside of available range",
+                       paste(range(1-as.numeric(dntab[[3]])),
+                             collapse="-")))
+
+        computeEmpProc <- function(X, y)
+        {
+               e <- as.vector(y - X %*% histcoef)
+               process <- rep(0, nrow(X)-K+1)
+               for(i in 0:(nrow(X)-K))
+               {
+                   process[i+1] <- sum(e[(i+1):(i+K)])
+               }
+               process/(sigmahat*sqrt(histsize))
+        }
+        functional <- "max"
+        computeEstims <- NULL
+        if(is.null(border)){
+            border <- function(k){
+                critval*sqrt(2*logPlus(k/histsize))
+            }
+        }
+    },
+
+    "ME" = {
+        if(is.null(CritvalTable))
+            CritvalTable <- get("monitorMECritvalTable")
+        dntab <- dimnames(CritvalTable)
+        if(!(winsize %in% dntab[[1]]))
+            stop(paste("winsize h =",winsize,"not available, we have:",
+                       paste(dntab[[1]], collapse=", ")))
+        if(!(period %in% dntab[[2]]))
+            stop(paste("period",period,"not available, we have:",
+                       paste(dntab[[2]], collapse=", ")))
+        critval <- approx(x=as.numeric(dntab[[3]]),
+                          y=CritvalTable[as.character(winsize),
                           as.character(period),,functional],
                           xout=1-elemsiglevel)$y
         if(is.na(critval))
@@ -85,21 +162,48 @@ mefp.efp <-
                 retval$Qr12 <- root.matrix(crossprod(x[ok,,drop=FALSE]))/sqrt(K)
             retval
         }
-        border <- function(k){
-            critval*sqrt(2*logPlus(k/histsize))
-        }
-    }
-    else{
 
-        mreSize <- function(a){
-            -2*(pnorm(a)-a*dnorm(a))
+        computeEmpProc <- function(newcoef, Q, k){
+            if(is.null(Q)) Q <- Q12
+            Q <- Q * K/(sigmahat*sqrt(histsize))
+            t(Q %*%(newcoef-histcoef))
         }
-        mreCritval <- function(a){
-            abs(2*(pnorm(a)-a*dnorm(a))+elemsiglevel-2)
+
+        if(is.null(border)){
+            border <- function(k){
+                critval*sqrt(2*logPlus(k/histsize))
+            }
+        }
+    },
+
+    "fluctuation" = {
+
+        if(is.null(CritvalTable)){
+            mreSize <- function(a){
+                -2*(pnorm(a)-a*dnorm(a))
+            }
+            mreCritval <- function(a){
+                abs(2*(pnorm(a)-a*dnorm(a))+elemsiglevel-2)
+            }
+            critval <- optim(5, mreCritval)$par
+            if((mreSize(critval)-elemsiglevel) > tolerance)
+                stop("Could not find critical within tolerance")
+        }
+        else{
+            dntab <- dimnames(CritvalTable)
+            if(!(period %in% dntab[[1]]))
+                stop(paste("period",period,"not available, we have:",
+                           paste(dntab[[1]], collapse=", ")))
+            critval <- approx(x=as.numeric(dntab[[2]]),
+                              y=CritvalTable[as.character(period),],
+                              xout=1-elemsiglevel)$y
+            if(is.na(critval))
+                stop(paste("Necessary significance level per parameter of",
+                           elemsiglevel,
+                           "\n\toutside of available range",
+                           paste(range(1-as.numeric(dntab[[3]])),
+                                 collapse="-")))
         }
-        critval <- optim(5, mreCritval)$par
-        if((mreSize(critval)-elemsiglevel) > tolerance)
-            stop("Could not find critical within tolerance")
 
 
         computeEstims <- function(x, y, k){
@@ -109,11 +213,20 @@ mefp.efp <-
                 retval$Qr12 <- root.matrix(crossprod(x[1:k,,drop=FALSE]))/sqrt(k)
             retval
         }
-        border <- function(k){
-            x <- k/histsize
-            sqrt(x*(x-1)*(critval^2 + log(x/(x-1))))
+
+        computeEmpProc <- function(newcoef, Q, k){
+            if(is.null(Q)) Q <- Q12
+            Q <- Q * k/(sigmahat*sqrt(histsize))
+            t(Q %*%(newcoef-histcoef))
         }
-    }
+
+        if(is.null(border)){
+            border <- function(k){
+                x <- k/histsize
+                sqrt(x*(x-1)*(critval^2 + log(x/(x-1))))
+            }
+        }
+    })
 
     if(functional=="max"){
         computeStat <- function(empproc){
@@ -130,8 +243,6 @@ mefp.efp <-
         }
     }
 
-
-
     obj <- list(breakpoint=NA, last=obj$nobs, process=NULL,
                 statistic=NULL, histsize=histsize,
                 initcall=match.call(), call=match.call(),
@@ -142,7 +253,7 @@ mefp.efp <-
                 functional=functional, alpha=alpha, critval=critval,
                 histcoef=histcoef, formula=obj$formula,
                 type.name=paste("Monitoring with", obj$type.name),
-                data=data, histtsp=obj$datatsp)
+                type=obj$type, data=data, histtsp=obj$datatsp)
 
     class(obj) <- "mefp"
     obj
@@ -162,7 +273,8 @@ monitor <- function(obj, data=NULL, verbose=TRUE){
 
     mf <- model.frame(obj$formula, data=data)
     y <- as.matrix(model.response(mf))
-    x <- model.matrix(obj$formula, data = data)
+    modelterms <- terms(obj$formula, data = data)
+    x <- model.matrix(modelterms, data = data)
 
     if(nrow(x) <= obj$last) return(obj)
     if(nrow(x)!=nrow(y))
@@ -170,40 +282,64 @@ monitor <- function(obj, data=NULL, verbose=TRUE){
     if(ncol(y)!=1)
         stop("multivariate response not implemented yet")
     foundBreak <- FALSE
-    for(k in (obj$last+1):nrow(x)){
-        newestims <- obj$computeEstims(x,y,k)
-        obj$process <- rbind(obj$process,
-                             obj$computeEmpProc(newestims$coef, newestims$Qr12))
-        stat <- obj$computeStat(obj$process)
-        obj$statistic <- c(obj$statistic, stat)
-        if(!foundBreak & (stat > obj$border(k))){
-            foundBreak <- TRUE
-            obj$breakpoint <- k
-            if(verbose) cat("Break detected at observation #", k, "\n")
-        }
+
+    if((obj$type == "OLS-MOSUM") | (obj$type == "OLS-CUSUM"))
+    {
+      if(obj$type == "OLS-CUSUM")
+      {
+        obj$process <- obj$computeEmpProc(x,y)[-(1:obj$histsize)]
+      }
+      else
+      {
+        obj$process <- obj$computeEmpProc(x,y)[-(1:length(obj$efpprocess))]
+      }
+      boundary <- obj$border((obj$histsize+1):nrow(x))
+      obj$statistic <- max(abs(obj$process))
+      if(!foundBreak & any(abs(obj$process) > boundary))
+      {
+        foundBreak <- TRUE
+        obj$breakpoint <- min(which(abs(obj$process) > boundary)) + obj$histsize
+        if(verbose) cat("Break detected at observation #", obj$breakpoint, "\n")
+      }
+      obj$lastcoef <- NULL
     }
-    obj$last <- k
-    obj$lastcoef <- newestims$coef
+    else
+    {
+      for(k in (obj$last+1):nrow(x)){
+          newestims <- obj$computeEstims(x,y,k)
+          obj$process <- rbind(obj$process,
+                               obj$computeEmpProc(newestims$coef, newestims$Qr12,k))
+          stat <- obj$computeStat(obj$process)
+          obj$statistic <- c(obj$statistic, stat)
+          if(!foundBreak & (stat > obj$border(k))){
+              foundBreak <- TRUE
+              obj$breakpoint <- k
+              if(verbose) cat("Break detected at observation #", k, "\n")
+          }
+      }
+      obj$lastcoef <- newestims$coef
+    }
+    obj$last <- nrow(x)
     obj$call <- match.call()
     obj
 }
 
-print.mefp <- function(obj, ...){
+print.mefp <- function(x, ...){
 
-    cat(obj$type.name, "\n\n")
-    cat("Initial call:\n ", deparse(obj$initcall), "\n\n")
-    cat("Last call:\n ", deparse(obj$call), "\n\n")
-    cat("Significance level   : ", obj$alpha, "\n")
-    cat("Critical value       : ", obj$critval, "\n")
-    cat("History size         : ", obj$histsize, "\n")
-    cat("Last point evaluated : ", obj$last, "\n")
-    if(!is.na(obj$breakpoint))
-        cat("Structural break at  : ", obj$breakpoint, "\n")
+    cat(x$type.name, "\n\n")
+    cat("Initial call:\n ", deparse(x$initcall), "\n\n")
+    cat("Last call:\n ", deparse(x$call), "\n\n")
+    cat("Significance level   : ", x$alpha, "\n")
+    cat("Critical value       : ", x$critval, "\n")
+    cat("History size         : ", x$histsize, "\n")
+    cat("Last point evaluated : ", x$last, "\n")
+    if(!is.na(x$breakpoint))
+        cat("Structural break at  : ", x$breakpoint, "\n")
     cat("\nParameter estimate on history :\n");
-    print(obj$histcoef)
-    if(!is.null(obj$lastcoef)){
+    print(x$histcoef)
+    if(!is.null(x$lastcoef)){
         cat("Last parameter estimate :\n");
-        print(obj$lastcoef)
+        print(x$lastcoef)
     }
 }
 
@@ -215,7 +351,7 @@ plot.mefp <- function(x, boundary=TRUE, functional="max", main=NULL,
                     as.matrix(x$process))
         proc <- ts(proc,
                    end=x$histtsp[2]+(x$last-x$histsize)/x$histtsp[3],
-                   frequency=frequency(x$efpprocess))
+                   frequency=x$histtsp[3])
         bound <- ts(x$border((x$histsize+1):x$last),
                  end = end(proc), frequency=frequency(proc))
         pos <- FALSE
